/*  $Id: StateMachine.java 1335 2012-08-25 13:27:49Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2002-2012 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.util.Hashtable;

	/** Automate à état fini 
     * <p>Automate générique. Le comportement de l'automate et la maintenance
     * de son état interne sont défini par des transitions. Chaque transition
     * est une règle qui précise depuis un état origine donnée, l'état cible à
     * atteindre. Pour autoriser la bascule d'un état à un autre il est donc 
     * nécessaire que l'automate possède la transition qui le décrit. Dans le
     * cas contraire la transition est refusée.</p>
	 * <p>ST désigne le type qui modélise un état (en général un enum ou un 
	 * int).</p>
     * <p>ET désigne le type qui modélise un événement, c'est à dire le 
     * phénomène qui induit un changement d'état (par exemple pour un
     * interpréteur de texte, il pourra s'agir de classes de caractères).</p>
     */ 
	public class StateMachine<ST,ET> implements Observable<StateMachine<ST,ET>>
	{
		/** Creation automate 
		 * @param startState Etat initial de l'automate
		 */ 
		public StateMachine(ST startState)
		{
			_state=startState;
			_transitions=new Hashtable<ST,Hashtable<ET,ST>>();
		}
		/** Dernier evenement à l'origine d'une transition. */
		private ET _lastEvent;
		/** Obtenir l'événement qui a provoqué la dernière transition
		 * <p>Attention, la valeur retourné n'est pas définie tant qu'aucune 
		 * transition n'a eu lieu.</p>
		 * @return 
		 * L'évenement qui a permis d'atteindre l'état courant.
		 */
		public ET getLastEvent()
		{
			return _lastEvent;
		}
		/**
		 * 
		 */
		public boolean isTransitionAllowed(ET ev) {
			boolean res=false;
			Hashtable<ET,ST> table=_transitions.get(_state);
			if (table!=null) {
				res=table.get(ev)!=null;
			}
			return res;
		}
		/**
		 *
		 */
		public Iterable<ET> getAllowedTransition() {
			Iterable<ET> res=null;
			Hashtable<ET,ST> table=_transitions.get(_state);
			if (table!=null) {
				res=table.keySet();
			}
			return res;
		}
		/**
		 * @param ev Evenement à soummettre à l'automate
		 * @return true si l'evenement est accepté, false si l'événement
		 * est incompatible avec l'état courant.
		 */ 
		public boolean runTransition(ET ev){
			boolean res=false;
			Hashtable<ET,ST> table=_transitions.get(_state);
			if (table!=null)
			{
				ST state=table.get(ev);
				if (state!=null)
				{
                    _state=state;
                    _lastEvent=ev;
                    _obsManager.notify(this);
					res=true;
				}
			}
			else
			{
				App.get().log().warn("CStateMachine in terminal state recieved event: "
							+ev+" [state="+_state+"]");
			}
			if (jc.O.DEBUG) {
				if (res) {
					App.get().log().debug("CStateMachine transition OK: "
								+ev+" [state="+_state+"]");
				}
				else {
					App.get().log().warn("CStateMachine transition refused: "
								+ev+" [state="+_state+"]");
				}
			}
			return res;
		}
		/** Ajouter une transition au graphe d'états de
		 * l'automate
		 * @param from Etat origine
		 * @param ev Evenement associé à la transition
		 * @param to Etat cible
		 */ 
		public void addTransition(ST from,ET ev,ST to)
		{
			Hashtable<ET,ST> table=_transitions.get(from);
			if (table==null) {
				table=new Hashtable<ET,ST>();
				_transitions.put(from,table);
			}
			table.put(ev,to);
		}
		/** Etat courant de l'automate */
		private ST _state;
		/** table des transitions valides 
		 * <p>dico indexé par les états de dico indexé par les evenements</p>
		 */ 
		private Hashtable<ST,Hashtable<ET,ST>> _transitions;
		 /** @return Etat courant de l'automate
		 */
		public ST getState()
		{
			return _state;
		}

		//---------------------------------------------------------------------
		/** Gestionnaire observateurs. */
        private ObserverManager<StateMachine<ST,ET>> _obsManager=
					new ObserverManager<StateMachine<ST,ET>>();
		/** @return Gestionnaire observateurs. */
        public ObserverManager<StateMachine<ST,ET>> getObserverManager()
        {
            return _obsManager;
        }
	}
